@@ -34,6 +34,7 @@
 #include "fr-command-zip.h"
 
 #define EMPTY_ARCHIVE_WARNING        "Empty zipfile."
+#define ZIP_SPECIAL_CHARACTERS       "[]*?!^-\\"
 
 static void fr_command_zip_class_init  (FrCommandZipClass *class);
 static void fr_command_zip_init        (FrCommand         *afile);
@@ -91,53 +92,6 @@ mktime_from_string (char *datetime_s)
 }
 
 
-static char *
-fr_command_zip_escape (FrCommand     *comm,
-		       const char    *str)
-{
-	char *estr;
-	char *estr2;
-
-	estr = escape_str (str, "\\");
-	estr2 = shell_escape (estr);
-
-	g_free (estr);
-
-	return estr2;
-
-}
-
-
-static char*
-zip_escape (const char *str)
-{
-	return escape_str (str, "*?[]");
-}
-
-
-static char*
-prepend_path_separator (const char *str)
-{
-	if (*str == '-' || g_str_has_prefix (str, "\\-"))
-		return g_strconcat (".", G_DIR_SEPARATOR_S, str, NULL);
-	else
-		return g_strdup (str);
-}
-
-
-static char*
-prepend_path_separator_zip_escape (const char *str)
-{
-	char *tmp1, *tmp2;
-
-	tmp2 = prepend_path_separator (str);
-	tmp1 = zip_escape (tmp2);
-	g_free (tmp2);
-
-	return tmp1;
-}
-
-
 static void
 list__process_line (char     *line,
 		    gpointer  data)
@@ -209,51 +163,34 @@ list__process_line (char     *line,
 
 
 static void
-add_filename_arg (FrCommand *comm)
+add_password_arg (FrCommand  *comm,
+		  const char *password)
 {
-	char *temp = prepend_path_separator (comm->e_filename);
-	fr_process_add_arg (comm->process, temp);
-	g_free (temp);
-}
-
-
-static void
-add_password_arg (FrCommand     *comm,
-		  const char    *password,
-		  gboolean       always_specify)
-{
-	if (always_specify || ((password != NULL) && (password[0] != '\0'))) {
-		char *arg;
-		char *e_password;
-
+	if ((password != NULL) && (password[0] != '\0')) {
 		fr_process_add_arg (comm->process, "-P");
-
-		e_password = escape_str (password, "\"*?[]'`()$!;");
-		if (e_password != NULL) {
-			arg = g_strconcat ("\"", e_password, "\"", NULL);
-			g_free (e_password);
-		} else
-			arg = g_strdup ("\"\"");
-
-		fr_process_add_arg (comm->process, arg);
-		g_free (arg);
+		fr_process_add_arg (comm->process, password);
 	}
 }
 
 
 static void
-fr_command_zip_list (FrCommand  *comm,
-		     const char *password)
+list__begin (gpointer data)
 {
-	FR_COMMAND_ZIP (comm)->is_empty = FALSE;
+	FrCommandZip *comm = data;
+	
+	comm->is_empty = FALSE;
+}
 
-	fr_process_set_out_line_func (FR_COMMAND (comm)->process,
-				      list__process_line,
-				      comm);
+
+static void
+fr_command_zip_list (FrCommand  *comm)
+{
+	fr_process_set_out_line_func (comm->process, list__process_line, comm);
 
 	fr_process_begin_command (comm->process, "unzip");
+	fr_process_set_begin_func (comm->process, list__begin, comm);
 	fr_process_add_arg (comm->process, "-ZTs");
-	add_filename_arg (comm);
+	fr_process_add_arg (comm->process, comm->filename);
 	fr_process_end_command (comm->process);
 	fr_process_start (comm->process);
 }
@@ -279,11 +216,11 @@ process_line__common (char     *line,
 
 static void
 fr_command_zip_add (FrCommand     *comm,
+		    const char    *from_file,
 		    GList         *file_list,
 		    const char    *base_dir,
 		    gboolean       update,
-		    const char    *password,
-		    FRCompression  compression)
+		    gboolean       recursive)
 {
 	GList *scan;
 
@@ -302,9 +239,9 @@ fr_command_zip_add (FrCommand     *comm,
 	if (update)
 		fr_process_add_arg (comm->process, "-u");
 
-	add_password_arg (comm, password, FALSE);
+	add_password_arg (comm, comm->password);
 
-	switch (compression) {
+	switch (comm->compression) {
 	case FR_COMPRESSION_VERY_FAST:
 		fr_process_add_arg (comm->process, "-1"); break;
 	case FR_COMPRESSION_FAST:
@@ -315,21 +252,18 @@ fr_command_zip_add (FrCommand     *comm,
 		fr_process_add_arg (comm->process, "-9"); break;
 	}
 
-	add_filename_arg (comm);
-
-	for (scan = file_list; scan; scan = scan->next) {
-		char *temp = prepend_path_separator ((char*) scan->data);
-		fr_process_add_arg (comm->process, temp);
-		g_free (temp);
-	}
+	fr_process_add_arg (comm->process, comm->filename);
+	for (scan = file_list; scan; scan = scan->next)
+		fr_process_add_arg (comm->process, scan->data);
 
 	fr_process_end_command (comm->process);
 }
 
 
 static void
-fr_command_zip_delete (FrCommand *comm,
-		       GList     *file_list)
+fr_command_zip_delete (FrCommand  *comm,
+		       const char *from_file,
+		       GList      *file_list)
 {
 	GList *scan;
 
@@ -339,12 +273,14 @@ fr_command_zip_delete (FrCommand *comm,
 
 	fr_process_begin_command (comm->process, "zip");
 	fr_process_add_arg (comm->process, "-d");
-	add_filename_arg (comm);
 
+	fr_process_add_arg (comm->process, comm->filename);
 	for (scan = file_list; scan; scan = scan->next) {
-		char *temp = prepend_path_separator_zip_escape ((char*) scan->data);
-		fr_process_add_arg (comm->process, temp);
-		g_free (temp);
+		char *escaped;
+		
+ 		escaped = escape_str (scan->data, ZIP_SPECIAL_CHARACTERS);
+ 		fr_process_add_arg (comm->process, escaped);
+ 		g_free (escaped);
 	}
 
 	fr_process_end_command (comm->process);
@@ -353,12 +289,12 @@ fr_command_zip_delete (FrCommand *comm,
 
 static void
 fr_command_zip_extract (FrCommand  *comm,
+			const char *from_file,
 			GList      *file_list,
 			const char *dest_dir,
 			gboolean    overwrite,
 			gboolean    skip_older,
-			gboolean    junk_paths,
-			const char *password)
+			gboolean    junk_paths)
 {
 	GList *scan;
 
@@ -369,31 +305,26 @@ fr_command_zip_extract (FrCommand  *comm,
 	fr_process_begin_command (comm->process, "unzip");
 
 	if (dest_dir != NULL) {
-		char *e_dest_dir = fr_command_escape (comm, dest_dir);
 		fr_process_add_arg (comm->process, "-d");
-		fr_process_add_arg (comm->process, e_dest_dir);
-		g_free (e_dest_dir);
+		fr_process_add_arg (comm->process, dest_dir);
 	}
-
 	if (overwrite)
 		fr_process_add_arg (comm->process, "-o");
 	else
 		fr_process_add_arg (comm->process, "-n");
-
 	if (skip_older)
 		fr_process_add_arg (comm->process, "-u");
-
 	if (junk_paths)
 		fr_process_add_arg (comm->process, "-j");
+	add_password_arg (comm, comm->password);
 
-	add_password_arg (comm, password, TRUE);
-
-	add_filename_arg (comm);
-
+	fr_process_add_arg (comm->process, comm->filename);
 	for (scan = file_list; scan; scan = scan->next) {
-		char *temp = prepend_path_separator_zip_escape ((char*) scan->data);
-		fr_process_add_arg (comm->process, temp);
-		g_free (temp);
+		char *escaped;
+		
+ 		escaped = escape_str (scan->data, ZIP_SPECIAL_CHARACTERS);
+ 		fr_process_add_arg (comm->process, escaped);
+ 		g_free (escaped);
 	}
 
 	fr_process_end_command (comm->process);
@@ -401,20 +332,19 @@ fr_command_zip_extract (FrCommand  *comm,
 
 
 static void
-fr_command_zip_test (FrCommand   *comm,
-		     const char  *password)
+fr_command_zip_test (FrCommand   *comm)
 {
 	fr_process_begin_command (comm->process, "unzip");
 	fr_process_add_arg (comm->process, "-t");
-	add_password_arg (comm, password, TRUE);
-	add_filename_arg (comm);
+	add_password_arg (comm, comm->password);
+	fr_process_add_arg (comm->process, comm->filename);
 	fr_process_end_command (comm->process);
 }
 
 
 static void
 fr_command_zip_handle_error (FrCommand   *comm,
-			     FRProcError *error)
+			     FrProcError *error)
 {
 	if (error->type == FR_PROC_ERROR_COMMAND_ERROR) {
 		if (error->status <= 1)
@@ -424,15 +354,15 @@ fr_command_zip_handle_error (FrCommand   *comm,
 		else {
 			GList *output;
 			GList *scan;
-			
+
 			if (comm->action == FR_ACTION_TESTING_ARCHIVE)
-				output = comm->process->raw_output;
+				output = comm->process->out.raw;
 			else
-				output = comm->process->raw_error;
-			
+				output = comm->process->err.raw;
+
 			for (scan = g_list_last (output); scan; scan = scan->prev) {
 				char *line = scan->data;
-				
+
 				if (strstr (line, "incorrect password") != NULL) {
 					error->type = FR_PROC_ERROR_ASK_PASSWORD;
 					break;
@@ -443,6 +373,39 @@ fr_command_zip_handle_error (FrCommand   *comm,
 }
 
 
+const char *zip_mime_type[] = { "application/x-cbz",
+				"application/x-ms-dos-executable",
+				"application/zip",
+				NULL };
+
+
+const char **
+fr_command_zip_get_mime_types (FrCommand *comm)
+{
+	return zip_mime_type;
+}
+
+
+FrCommandCap
+fr_command_zip_get_capabilities (FrCommand  *comm,
+			         const char *mime_type)
+{
+	FrCommandCap capabilities;
+
+	capabilities = FR_COMMAND_CAN_ARCHIVE_MANY_FILES | FR_COMMAND_CAN_ENCRYPT;
+	if (is_program_in_path ("zip")) {
+		if (strcmp (mime_type, "application/x-ms-dos-executable") == 0)
+			capabilities |= FR_COMMAND_CAN_READ;
+		else
+			capabilities |= FR_COMMAND_CAN_READ_WRITE;
+	}
+	else if (is_program_in_path ("unzip"))
+		capabilities |= FR_COMMAND_CAN_READ;
+
+	return capabilities;
+}
+
+
 static void
 fr_command_zip_class_init (FrCommandZipClass *class)
 {
@@ -454,22 +417,20 @@ fr_command_zip_class_init (FrCommandZipClass *class)
 
 	gobject_class->finalize = fr_command_zip_finalize;
 
-	afc->list           = fr_command_zip_list;
-	afc->add            = fr_command_zip_add;
-	afc->delete         = fr_command_zip_delete;
-	afc->extract        = fr_command_zip_extract;
-	afc->test           = fr_command_zip_test;
-	afc->handle_error   = fr_command_zip_handle_error;
-
-	afc->escape         = fr_command_zip_escape;
+	afc->list             = fr_command_zip_list;
+	afc->add              = fr_command_zip_add;
+	afc->delete           = fr_command_zip_delete;
+	afc->extract          = fr_command_zip_extract;
+	afc->test             = fr_command_zip_test;
+	afc->handle_error     = fr_command_zip_handle_error;
+	afc->get_mime_types   = fr_command_zip_get_mime_types;
+	afc->get_capabilities = fr_command_zip_get_capabilities;
 }
 
 
 static void
 fr_command_zip_init (FrCommand *comm)
 {
-	comm->file_type = FR_FILE_TYPE_ZIP;
-
 	comm->propAddCanUpdate             = TRUE;
 	comm->propAddCanReplace            = TRUE;
 	comm->propAddCanStoreFolders       = TRUE;
@@ -521,21 +482,3 @@ fr_command_zip_get_type ()
 
 	return type;
 }
-
-
-FrCommand *
-fr_command_zip_new (FrProcess  *process,
-		    const char *filename)
-{
-	FrCommand *comm;
-
-	if (!is_program_in_path ("zip") &&
-	    !is_program_in_path ("unzip")) {
-		return NULL;
-	}
-
-	comm = FR_COMMAND (g_object_new (FR_TYPE_COMMAND_ZIP, NULL));
-	fr_command_construct (comm, process, filename);
-
-	return comm;
-}
